package HFE;

import java.util.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing QuartzLight in Java
//-----------------------------------------------------------------------------

/**
<p>
This class provides a vector of polynomials of degree one in n independent variables.
The polynomials are over the finite field GF(2) each. As all these
polynomials are processed in parallel, the coefficients
are represented in terms of {@link Field2_n} each.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.15
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


final class AffinePolyVec {

    /** The number of variables in this polynomial.
      @serial is an integer
    */
    public static int varNum = HFE2_n.extensionDegree;

   /** The number of rows in this vector.
      @serial is an integer
    */
    public static int elemNum = HFE2_n.extensionDegree;

    /** Coefficients for the variables.
        The variables are x_1..x_n where
        n is {@link #varNum}. Position 0
        is for the constant terms.
    */
    private Field2_n coeff[];


    /** Initializes all coefficients to zero. The resulting polynomial vector is _not_ affine!
    */
    public AffinePolyVec() {
      //Check.assert(varNum >= 0);
      initMem();
    }


    /** Creates a copy from the given polynomial.
        Uses memory copy for the coefficients.
    */
    public AffinePolyVec(final AffinePolyVec father) {
      initMem();
      for (int i = 0; i <= varNum; i++)
        coeff[i].copy(father.coeff[i]);
    }


    /** {@link #transform2poly}
    */
    public AffinePolyVec(final Transform2 inTransform) {
      initMem();
      int[] m = inTransform.m;
      for (int i = 0; i < elemNum; i++)
        if (m[i] != 0) coeff[0].setBit(i);
      int[][] Ms = inTransform.M.getArray();
      for (int i = 0; i < elemNum; i++) {
        for (int j = 0; j < elemNum; j++)
          if (Ms[j][i] != 0) coeff[i+1].setBit(j);
      }
    }

    /** Returns the coefficients for a given variable.
        Valid values: 0: constant term, 1..valNum: variable
    */
    public Field2_n coeff(int which) {
      //Check.assert((which >= 0) && (which <= varNum));
      return coeff[which];
    }

    /** returns the coefficient array
        without copying it.
    */
    public Field2_n[] coeffs() {
      return coeff;
    }

    /** Compares this and whom coefficientwise.
        If the two polynomials have different number of variables they
        can not be equal.
      @return true iff all the coefficients are the same
              both objects
    */
    public boolean isEqual(AffinePolyVec whom) {
      for (int i = 0; i <= varNum; i++) {
        if (!whom.coeff[i].isEqual(coeff[i])) return false;
      }
      return true;
    }


    /** Sets all the coefficients to zero
    */
    public void setZero() {
      for (int i = 0; i <= varNum; i++)
        coeff[i].setZero();
    }


    /** Allocates new memory for the coefficients.
        The coefficients are initalized with zero.
    */
    public void initMem() {
      coeff = new Field2_n[varNum+1];
      for (int i = 0; i <= varNum; i++)
        coeff[i] = new Field2_n();
    }

    /** Initializes all coefficients with random bits
    */
    public void rnd(Random rnd) {
      for (int i = 0; i <= varNum; i++)
        coeff[i].rnd(rnd);
    }

    /** Sets all the coefficients
        by the given array. Does neither memory
        copy nor argument checking.
    */
    public void setCoeffs(Field2_n [] inArray) {
      //Check.assert(varNum == (inArray.length-1));
      coeff = inArray;
    }


    /** Transforms an affine transformation into
        an affine polynomial vector.
       @param inTransform affine tranformation
    */
    public void transform2poly(Transform2 inTransform) {
      // start with the constant term
      coeff[0].setZero();
      int[] m = inTransform.m;
      for (int i = 0; i < elemNum; i++)
        if (m[i] != 0) coeff[0].setBit(i);
      int[][] Ms = inTransform.M.getArray();
      for (int i = 0; i < elemNum; i++) {
        coeff[i+1].setZero();
        for (int j = 0; j < elemNum; j++)
          if (Ms[j][i] != 0) coeff[i+1].setBit(j);
      }
    }

    /** calculates this = a + b
        as polynomial addition
    */
    public void add(final AffinePolyVec a, final AffinePolyVec b) {
      for (int i = 0; i <= varNum; i++)
        coeff[i].add(a.coeff[i],b.coeff[i]);
    }


    /** calculate this += whom
        as polynomial addition
    */
    public void add(final AffinePolyVec whom) {
      for (int i = 0; i <= varNum; i++)
        coeff[i].add(whom.coeff[i]);
    }

    /** calculates this = a - b
        as polynomial subtraction
    */
    public void sub(final AffinePolyVec a, final AffinePolyVec b) {
      for (int i = 0; i <= varNum; i++)
        coeff[i].sub(a.coeff[i],b.coeff[i]);
    }


    /** calculate this -= whom
        as polynomial subtraction
    */
    public void sub(final AffinePolyVec whom) {
      for (int i = 0; i <= varNum; i++)
        coeff[i].sub(whom.coeff[i]);
    }

    /** Evaluate the polynomial using the given value vector
        The entry value[0] is _not_ uses, value[1] is the first valid entry.
      @param res value of the current polynimial vector.
      @param values array with varNum+1 entries
    */
    public void apply(Field2_n res, int[] values) {
      //Check.assert(values.length == (varNum+1));
      //Check.assert(Field2.isElement(values));
      res.copy(coeff[0]);
      for (int i = 1; i <= varNum; i++)
        if (values[i] == 1) res.add(coeff[i]);
    }

    /** Evaluate the polynomial using the given value vector
      @param res value of the current polynimial vector.
      @param value array with varNum entries as {@link Field2_n} element
    */
    public void apply(Field2_n res, Field2_n value) {
      res.copy(coeff[0]);
      for (int i = 1; i <= varNum; i++)
        if (value.checkBit(i-1)) res.add(coeff[i]);
    }

    /** Evaluate the polynomial using the given value vector.
        Does _not_ apply the constant terms
      @param res value of the current polynimial vector.
      @param value array with varNum entries as {@link Field2_n} element
    */
    public void applyLinear(Field2_n res, Field2_n value) {
      res.setZero();
      for (int i = 1; i <= varNum; i++)
        if (value.checkBit(i-1)) res.add(coeff[i]);
    }

    /** Computes an affine polynomial vector which is exactly
        a squaring of the original one.
    */
    public AffinePolyVec sqr() {
      AffinePolyVec res = new AffinePolyVec();
      for (int i = 0; i <= varNum; i++)
        res.coeff[i].sqr(coeff[i]);
      return res;
    }

    /** res = fac*this
        in terms of affine polynomial vector multiplication
    */
    public void mul(AffinePolyVec res, Field2_n fac) {
      Field2_n[] rc = res.coeff;
      for (int i = 0; i <= varNum; i++)
        rc[i].mul(fac,coeff[i]);
    }

    /** Writes the polynomials as a string to stdout
    */
    public void writeString() {
      for (int i = 1; i <= varNum; i++)
        if (!coeff[i].isZero()) System.out.print(coeff[i].stringIt()+ " x_" + i + " + ");
      System.out.print(coeff[0].stringIt());
    }


    /** Tests the functionality for the whole class.
        Reports errors to stdout.
    */
    public static void testIt () {
      System.out.println("start testing AffinePolyVec ...");

      final int repeats = 100;

      Random rnd = new Random();
      AffinePolyVec cur = new AffinePolyVec(), cur1 = new AffinePolyVec(),
                       cur2 = new AffinePolyVec(), cur3 = new AffinePolyVec();
      Transform2 trans1 = new Transform2(rnd), trans2 = new Transform2(rnd);
      int curRnd, curRnd1, curRnd2, curRnd3, curRes;
      int [] testVec = new int[varNum], outVec = new int[varNum], valueVec = new int[varNum+1];
      Field2_n out1 = new Field2_n(), out2 = new Field2_n(), out3 = new Field2_n(), in1 = new Field2_n(),
               in2 = new Field2_n();

      for (int i = 0; i < repeats; i++) {
        for (int j = 0; j < testVec.length; j++) {
          testVec[j] = Field2.rnd(rnd);
          valueVec[j+1] = testVec[j];
        }
        in1.setElement(testVec);
        cur1 = new AffinePolyVec(trans1);
        cur2 = new AffinePolyVec(cur1);
        cur1.apply(out1,valueVec);
        cur2.apply(out2,in1);
        if (!out2.isEqual(out1)) System.out.println("Error 1.1 with " + out1.stringIt() + " and " + out2.stringIt());
        trans1.transform(testVec,outVec);
        out2.setElement(outVec);
        if (!out2.isEqual(out1)) System.out.println("Error 1.2 with " + out1.stringIt() + " and " + out2.stringIt());

      }

      // test addition
      for (int i = 0; i < repeats; i++) {
        in1.rnd(rnd);
        cur1 = new AffinePolyVec(trans1);
        cur2 = new AffinePolyVec(trans2);
        cur.add(cur1,cur2);  // addition of two vectors should have the same result as adding the two results

        cur1.apply(out1,in1);
        cur2.apply(out2,in1);
        cur.apply(out3,in1);
        in1.add(out1,out2);
        if (!in1.isEqual(out3)) System.out.println("Erro 1.3 with "+ in1.stringIt() + " and " + out3.stringIt());
      }

      // test squaring
      for (int i = 0; i < repeats; i++) {
        in1.rnd(rnd);
        cur1 = new AffinePolyVec(trans1);
        cur2 = cur1.sqr();
        cur1.apply(out1,in1);
        out2.sqr(out1);
        cur2.apply(out3,in1);

        if (!out2.isEqual(out3)) System.out.println("Erro 1.4 with "+ out2.stringIt() + " and " + out3.stringIt());
      }

      // test multiplication
      for (int i = 0; i < repeats; i++) {
        in1.rnd(rnd); in2.rnd(rnd);
        cur1 = new AffinePolyVec(trans1);
        cur1.mul(cur2,in2);
        cur1.apply(out1,in1);
        cur2.apply(out2,in1);
        out3.mul(out1,in2);

        if (!out2.isEqual(out3)) System.out.println("Erro 1.5 with "+ out2.stringIt() + " and " + out3.stringIt());
      }


      System.out.println("...finished testing.");
    }

    /** Mainly invoces the {@link #testIt} method.
    */
    public static void main(String[] args) {
        testIt();
    }
}
