package HFE;

import java.lang.*;
import java.io.*;
import java.util.*;
import java.security.*;
import sun.misc.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing QuartzLight in Java
//-----------------------------------------------------------------------------

/**
<p>
This class is the "central class" of QuartzLight and is
used for speed measurement.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.15
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


class HFE2_n implements CheckRoot {

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// Central switches to select E

  /** Central switch to change generating polynomial.
      Saves only the powers as the coefficients must be 1.
      The same polynomial (but in a different format) is stored in
      {@link #generatingPolynomialAsMaxPoly}
  */
  final static int[] generatingPolynomial = new int[] {103, 9, 0}; // f(x) = x^103 + x^9 + 1

  /** Central switch to change the reduction term,
     i.e. the part of the polynomial with is not the degree and not zero
  */
  final static int reductionTerm = 513;   //  f(x) = x^67 + x^9 + 1 (i.e. 513 = 512 + 1 as reduction term)

  /** Holds the degree of the irreducible polynomial and thereby also
      the degree of the extension field
  */
  final static int extensionDegree = 103;

//!!!!!!!!!!!!!!!!!!!!!!!!!!!

  /** Polynomial set in variable {@link #generatingPolynomial}
      but as a Poly2Max polynomial.
      Does not need to be changed for different {@link Field2_n}.
  */
  final static Poly2Max generatingPolynomialAsMaxPoly = new Poly2Max(generatingPolynomial);

  /** Holds the number of bits for a message in terms of a signature.
      Is 7 in Quartz.
  */
  final static int padBits = 7;

  /** Holds the number of bits for a message in terms of a signature.
      This is smaller than messageLength as we need some "extra bits" for
      random padding.
  */
  final static int messageSigLength = extensionDegree - padBits;

  /** Holds the number of bits for a signature.

  */
  final static int signatureLength = extensionDegree;

  /** signals if we found a root or not, i.e., if the root-field is valid or not.
  */
  boolean foundRoot = false;

  /** Stores the private polynomial over GF(2^n). Is used for decryption.
    @serial Private Key
  */
  private Poly2_nMax privatePoly;

  /** Stores the public key in an array of polynomials.
      Is used for encryption - slow variation!
    @serial Poly2MultiSquare
  */
  private Poly2MultiSquare[] publicPoly;

  /** Transformation S in HFE
  */
  private Transform2 transS;

  /** Transformation T in HFE
  */
  private Transform2 transT;

  /** Transform S in HFE in terms of affine polynomials
  */
  private AffinePolyVec transSpoly;

  /** Transform T in HFE in terms of affine polynomials
  */
  private AffinePolyVec transTpoly;

  /** Root finding algorithm
  */
  private RootFinding rootFinder = new RootFinding();

  /** Message as a vector. For the sake of speed is it a static.
    @serial vector
  */
  private static int[] textVec = new int[extensionDegree];

  /** Hash of message as a vector. For the sake of speed is it a static.
      Is used in {@link #genSign}
    @serial vector
  */
  private static int[] sigTmpVec = new int[extensionDegree];

  /** Message as a vector. For the sake of speed is it a static.
    @serial vector
  */
  private static int[] resVec = new int[extensionDegree];

  /** Message as a vector. For the sake of speed is it a static.
      Is used in {@link #genSign}
    @serial vector
  */
  private static int[] rootVec = new int[extensionDegree];

  /** Message in a special format as a vector. For the sake of speed is it a static.
    @serial vector
  */
  private static int[] valueVec = new int[extensionDegree+1];

  /** Cipher as a vector. For the sake of speed static.
  */
  private static int[] codeVec = new int[Poly2MultiSquare.tau(extensionDegree)];

  /** Temporary register [element of GF(2^n)]. Static for the sake of speed
  */
  private static Field2_n tmpElem1 = new Field2_n();

  /** Temporary register [element of GF(2^n)]. Static for the sake of speed
  */
  private static Field2_n tmpElem2 = new Field2_n();

  /** Random number generator for signature generation
  */
  Random rnd = new SecureRandom();


  /** Initializes HFE2_n by generating the
      public and the private key from the given seed. Also generates the
      message digest.
     @param seed Seed for the cryptographic random number generator.
  */
  HFE2_n (byte[] seed) {
    publicPoly = new Poly2MultiSquare[extensionDegree];
    for (int i = 0; i < extensionDegree; i++)
      publicPoly[i] = new Poly2MultiSquare(extensionDegree);
    generateKeys(seed);
  }


  /** Generates a private polynomial for HFE.
     @param  rndGen Secure random number generator
     @return private polynomial
  */
  private static Poly2_nMax genPolyPrivate(SecureRandom rndGen) {
    Poly2_nMax res = new Poly2_nMax();
    // set the leading coefficient to have value 1
    // this is _not_ as specified in the Quartz standard
    // however, as this implementation is HFE- rather
    // than HFEv-, this change is no problem
    res.setCoeff(new Field2_n(1), 129);
    // set the coefficients to random values
    for (int i = 0; i < 8; i++)
      for (int j = i+1; j < 8; j++) {
        int degree = (1<<i) + (1<<j);
        if (degree < 129) res.setCoeff(rndGen,degree);
      }
    for (int i = 0; i < 8; i++) {
      int degree = 1<<i;
      res.setCoeff(rndGen,degree);
    }
    return res;
  }


  /** Generats the public polynomials from the private polynomial
      using base transformation. This implementation is slow
      as it is based on slow implementations of multivariables
      polynomial vectors
     @return public polynomials
  */
  private Poly2MultiSquare[] genPolyPublicSlow() {

    int[] coeffVec = new int[extensionDegree];
    Field2_n coeff = new Field2_n();
    Multi2Power powers = new Multi2Power(generatingPolynomialAsMaxPoly, transS.getTransform(), 7);
    Multi2Mul calc = new Multi2Mul(generatingPolynomialAsMaxPoly);

    Poly2MultiSquare[] res = new Poly2MultiSquare[extensionDegree];
    for (int i = 0; i < extensionDegree; i++)
      res[i] = new Poly2MultiSquare(extensionDegree);

    Poly2MultiSquare[] resMulti = new Poly2MultiSquare[extensionDegree];
    for (int i = 0; i < extensionDegree; i++)
      resMulti[i] = new Poly2MultiSquare(extensionDegree);
    Poly2MultiLinear[] resLin = new Poly2MultiLinear[extensionDegree];
    for (int i = 0; i < extensionDegree; i++)
      resLin[i] = new Poly2MultiLinear(extensionDegree);

    // compute the square terms
    int[] squareTerms = new int []{7,0, 6,0, 5,0, 4,0, 3,0, 2,0, 1,0,
                                        6,1, 5,1, 4,1, 3,1, 2,1,
                                        6,2, 5,2, 4,2, 3,2,
                                        6,3, 5,3, 4,3,
                                        6,4, 5,4,
                                        6,5
                                        };
    for (int i = 0; i < squareTerms.length; i+= 2) {
      int i1 = squareTerms[i];
      int i2 = squareTerms[i+1];
      int pos = (1 << i1) + (1 << i2);
      privatePoly.getCoeff(coeff,pos);
      coeff.getElement(coeffVec);
      calc.mul(resMulti,coeffVec,powers.givePower(i1),powers.givePower(i2));
      for (int j = 0; j < extensionDegree; j++)
        res[j].add(resMulti[j]);
    }

    // compute the linear terms
    int[] linearTerms = new int []{7, 6, 5, 4, 3, 2, 1, 0};
    for (int i = 0; i < linearTerms.length; i++) {
      int i1 = linearTerms[i];
      int pos = 1 << i1;
      privatePoly.getCoeff(coeff,pos);
      coeff.getElement(coeffVec);
      calc.mul(resLin,coeffVec,powers.givePower(i1));
      for (int j = 0; j < extensionDegree; j++)
        res[j].add(resLin[j]);
    }

    // overwrite or result vector with the T transformation
    res = transT.transform(res);

    return res; // after "a few steps" we finally got the public polynomials
  }

  /** Generats the public polynomials from the private polynomial
      using base transformation.
     @return public polynomials
  */
  private void genPolyPublic() {
    int[] coeffVec = new int[extensionDegree];
    final int logD = 7; // binary logarithm of the degree of the private polynomial
    Field2_n coeff = new Field2_n();
    AffinePolyVec [] powers = new AffinePolyVec[logD+1];
    // start with the represantation of S
    powers[0] = new AffinePolyVec(transS);
    // compute the squares for each affine transformation
    for (int i = 1; i < powers.length; i++)
      powers[i] = powers[i-1].sqr();
    SquarePolyVec res = new SquarePolyVec();
    SquarePolyVec tmpSq = new SquarePolyVec();
    AffinePolyVec tmpAf = new AffinePolyVec();

    // compute the square terms
    int[] squareTerms = new int []{7,0, 6,0, 5,0, 4,0, 3,0, 2,0, 1,0,
                                        6,1, 5,1, 4,1, 3,1, 2,1,
                                        6,2, 5,2, 4,2, 3,2,
                                        6,3, 5,3, 4,3,
                                        6,4, 5,4,
                                        6,5
                                        };
    for (int i = 0; i < squareTerms.length; i+= 2) {
      int i1 = squareTerms[i];
      int i2 = squareTerms[i+1];
      int pos = (1 << i1) + (1 << i2);
      privatePoly.getCoeff(coeff,pos);
      tmpSq.mul(powers[i1],powers[i2]);
      tmpSq.mul(coeff);
      res.add(tmpSq);
    }
    // compute the linear terms
    int[] linearTerms = new int []{7, 6, 5, 4, 3, 2, 1, 0};
    //int[] linearTerms = new int []{1, 0};
    for (int i = 0; i < linearTerms.length; i++) {
      int i1 = linearTerms[i];
      int pos = 1 << i1;
      privatePoly.getCoeff(coeff,pos);
      powers[i1].mul(tmpAf,coeff);
      res.add(tmpAf);
    }
    // get result vector after appying affine transformation T
    Field2_n[] coeffs = res.coeffs();
    transTpoly.apply(coeff,coeffs[0]);
    coeffs[0].copy(coeff);
    for (int i = 1; i < coeffs.length; i++) {
      transTpoly.applyLinear(coeff,coeffs[i]);
      coeffs[i].copy(coeff);
    }
    for (int i = 0; i < extensionDegree; i++) {
      int[] intCoeff = publicPoly[i].coeffs();
      for (int j = 0; j < intCoeff.length; j++)
        if (coeffs[j].checkBit(i)) intCoeff[j] = 1; else intCoeff[j] = 0;
    }
  }

  /** Generats the public polynomials from the private polynomial
      using polynomial interpolation.
     @return public polynomials
  */
  private void genPolyPublicInterpoly(Poly2MultiSquare[] res) {
    // init the polynomials
    SquarePolyVec tmpPoly = new SquarePolyVec();
    Field2_n tmpElem = new Field2_n();
    Field2_n[] coeffs = tmpPoly.coeffs();
    // start with evaluating 0
    encryptPrivate(tmpElem,coeffs[0]);

    // now the linear terms
    for (int i = 0; i < extensionDegree; i++) {
      tmpElem.setBit(i);
      encryptPrivate(tmpElem,coeffs[i+1]);
      coeffs[i+1].sub(coeffs[0]);
      tmpElem.clearBit(i);
    }

    // last but not least the quadratic terms
    int k = extensionDegree+1;
    for (int i = 0; i < extensionDegree; i++) {
      tmpElem.setBit(i);
      for (int j = i+1; j < extensionDegree; j++) {
        tmpElem.setBit(j);
        encryptPrivate(tmpElem,coeffs[k]);
        coeffs[k].sub(coeffs[i+1]);
        coeffs[k].sub(coeffs[j+1]);
        coeffs[k].add(coeffs[0]);
        k++;
        tmpElem.clearBit(j);
      }
      tmpElem.clearBit(i);
    }
    for (int i = 0; i < extensionDegree; i++) {
      int[] coeff = res[i].coeffs();
      for (int j = 0; j < Poly2MultiSquare.tau(extensionDegree); j++) {
        if (coeffs[j].checkBit(i)) coeff[j] = 1; else coeff[j] = 0;
      }
    }
  }

  /** Generates the public and the private key from
      a given seed and stores them in the object's fields.
     @param seed Seed for the random number generator. Must be "quite long" to be secure
     (e.g., at least 80 bits in the Quartz standard).
  */
  void generateKeys(byte [] seed) {
    SecureRandom rndGen = new SecureRandom();
    rndGen.setSeed(seed);
    privatePoly = genPolyPrivate(rndGen);
    transS = new Transform2(rndGen);
    transSpoly = new AffinePolyVec(transS);  // we don't need it here, but for a fair comparision
    transT = new Transform2(rndGen);
    transTpoly = new AffinePolyVec(transT);
    genPolyPublic();
  }

  /** Generates the public and the private key from
      a given seed and stores them in the object's fields.
     @param seed Seed for the random number generator. Must be "quite long" to be secure
     (e.g., at least 80 bits in the Quartz standard).
  */
  void generateKeysInterpoly(byte [] seed) {
    SecureRandom rndGen = new SecureRandom();
    rndGen.setSeed(seed);
    privatePoly = genPolyPrivate(rndGen);
    transS = new Transform2(rndGen);
    transSpoly = new AffinePolyVec(transS);  // saves time, as this method is faster...
    transT = new Transform2(rndGen);
    transTpoly = new AffinePolyVec(transT);
    genPolyPublicInterpoly(publicPoly);
  }

  /** Encryptes a given, n bit plain text.
      <br>
      Uses <b>private</b> key. Is used during public key generation.
      Needs {@link #transSpoly}, {@link #transTpoly}
      to be initialized correctly.
     @param plain plaintext as element from {@link Field2_n}
     @param res result as element from {@link Field2_n}
  */
  private void encryptPrivate(Field2_n plain, Field2_n res) {
    // apply transformation S
    transSpoly.apply(tmpElem1,plain);
    privatePoly.applyQL(tmpElem2,tmpElem1);
    transTpoly.apply(res,tmpElem2);
  }


  /** Generates signature for given message using (S,P,T).
    @param message to sign (in general: hash of this message with {@link #messageSigLength} bits)
    @param res signature for this message with {@link #signatureLength} bits
  */
  public void genSign(int[] message, int[] res) {
    //Check.assert(message.length == messageSigLength, "message has wrong length: " + message.length);
    //Check.assert(res.length == signatureLength, "result has wrong length: " + res.length);

    // copy message and set the padding bits by random
    System.arraycopy(message,0,sigTmpVec,0,messageSigLength);
    for (int i = messageSigLength; i < signatureLength; i++)
      sigTmpVec[i] = Field2.rnd(rnd);

    for (int i = 0; i < (1 << padBits); i++) {
      foundRoot = false; // we don't have any root up to now
      System.arraycopy(sigTmpVec,0,rootVec,0,signatureLength);  // bad style, but anyway: rootVec it not needed in this step!
      transT.untransform(rootVec,resVec);

      // use root finding algorithm on element of GF(2^n)
      tmpElem1.setElement(resVec);
      privatePoly.setCoeff(tmpElem1,0);  // exploit the fact that coeff0 of privatePoly is zero, and also, that the leading term is 1
      rootFinder.findRoots(privatePoly,this);
      if (foundRoot) {  // did we get a root?
        // copy to res-vector which was set in checkRoot
        System.arraycopy(rootVec,0,res,0,signatureLength);
        return;
      }
      // check next padding
      boolean carry = true;
      int j = messageSigLength;
      while (carry && (j < signatureLength)) {
        sigTmpVec[j] ^= 1; // add one in GF(2)
        carry = (sigTmpVec[j] == 0);
        j++;
      }
    }
    // if we ever reach this part we are in trouble!
    // it's not the correct exception be we don't have to declare this one...
    throw new IllegalArgumentException("Did not find a suitable root for signature.");
  }

  /** Validates if given signature <code>sig</code> is valid or not.
    @param message which was signed (in general: hash of this message with {@link #messageSigLength} bits) using method {@link #genSign}
    @param sig signature for this message with {@link #signatureLength} bits
    @return true iff sig is valid for current public key
  */
  public boolean valSign(int[] message, int[] sig) {
    //Check.assert(message.length == messageSigLength, "message has wrong length: " + message.length);
    //Check.assert(sig.length == signatureLength, "signature has wrong length: " + sig.length);

    System.arraycopy(sig,0,valueVec,1,signatureLength);
    Poly2MultiSquare.getValues(valueVec,codeVec);
    for (int i = 0; i < messageSigLength; i++)
      if (message[i] != publicPoly[i].apply(codeVec)) return false;

    return true;
  }

  /** Collects the roots. In an encryption scheme, it would also check if the root is the correct message.
     @return false
  */
  public boolean checkRoot(Field2_n root) {
    // signature mode: we want _all_ roots
    // we don't use SHA-1 here, but only the smalles root
    root.getElement(textVec);
    transS.untransform(textVec,sigTmpVec);
    if (foundRoot == false) {
      foundRoot = true;
      // copy the root to the tmp-field
      System.arraycopy(sigTmpVec,0, rootVec,0,signatureLength);
      return false;
    }
    // check if the current root is smaller
    for (int i = sigTmpVec.length-1; i >= 0; i--) {
      if (sigTmpVec[i] > rootVec[i]) return false; // the new value is too big
      if (sigTmpVec[i] < rootVec[i]) i = -1;       // the new value is O.K.
    }
    System.arraycopy(sigTmpVec,0, rootVec,0,signatureLength);
    return false; // we want to get _any_ root
  }


  /** Tests the functionality of the whole class.
      If there occurs any error it will be printed on stdout.
  */
  public static void testIt() {
    System.out.println("Start testing HFE2_n...");
System.out.println("System parameter: ");
System.out.println("  Extension degree: " + extensionDegree);

    HFE2_n testObj = new HFE2_n(new byte[] {1, 2, 3, 4, 5, 6, 7, 8});
    testObj.transSpoly = new AffinePolyVec(testObj.transS);
    testObj.transTpoly = new AffinePolyVec(testObj.transT);

    Random rnd = new Random();
    int[] dec = new int[extensionDegree];
    int[] sig = new int[messageSigLength];
    Field2_n in1 = new Field2_n(), out1 = new Field2_n(), out2 = new Field2_n();
    Poly2MultiSquare[] polys2;

    Poly2MultiSquare[] publicPolys = new Poly2MultiSquare[extensionDegree];
    for (int i = 0; i < extensionDegree; i++)
      publicPolys[i] = new Poly2MultiSquare(extensionDegree);
    testObj.genPolyPublicInterpoly(publicPolys);
    for (int i = 0; i < 100; i++) {
      in1.rnd(rnd);
      testObj.encryptPrivate(in1,out1);

      in1.getElement(dec);
      System.arraycopy(dec,0,valueVec,1,extensionDegree);
      Poly2MultiSquare.getValues(valueVec,codeVec);
      for (int j = 0; j < extensionDegree; j++)
        dec[j] = publicPolys[j].apply(codeVec);
      out2.setElement(dec);
      if (!out1.isEqual(out2)) System.out.println("Error 1.1");
    }

    polys2 = testObj.genPolyPublicSlow();
    for (int i = 0; i < polys2.length; i++) {
      if (!polys2[i].isEqual(testObj.publicPoly[i])) System.out.println("Error 1.2");
    }

    for (int i = 0; i < 100; i++) {
      in1.rnd(rnd);
      testObj.encryptPrivate(in1,out1);

      in1.getElement(dec);
      System.arraycopy(dec,0,valueVec,1,extensionDegree);
      Poly2MultiSquare.getValues(valueVec,codeVec);
      for (int j = 0; j < extensionDegree; j++)
        dec[j] = testObj.publicPoly[j].apply(codeVec);
      out2.setElement(dec);
      if (!out1.isEqual(out2)) System.out.println("Error 1.3");
    }

    for (int i = 0; i < 100; i++) {
      System.out.println("Run " + i);

      for (int k = 0; k < sig.length; k++)
        sig[k] = Field2.rnd(rnd);

      testObj.genSign(sig,dec);
      if (!testObj.valSign(sig,dec)) System.out.println("Signature not valid!");
    }
    
    System.out.println("...ended testing.");
  }

  /** Tests how fast the package runs on the current hardware.
  */
  public final static void speedIt() {
    //Check.assert(false);  // make sure we do not have any assertion checkers
    long overallTime = System.currentTimeMillis();

    System.out.println("Start speed measurement HFE2_n...");
    System.out.print("System parameter: ");
    System.out.println(" Extension degree: " + extensionDegree);
    Random rnd = new Random();
    byte [] seed = new byte[] {1, 2, 3, 4, 5, 6, 7, 8};
    HFE2_n mySpeed = new HFE2_n(seed);
    long curTime;
    int paramM;
    int repeats;
    int[] plainVec = new int[extensionDegree];
    int[] encVec = new int[signatureLength];

    /**/
    System.gc();
    // key generation
    paramM = 102;
    System.out.println("n=" + paramM);
    System.out.println("r=1");
    System.out.println("Key Generation [QuartzLight, E=GF(2^" + extensionDegree +  ")]");
    for (int i = 0; i < paramM; i++) {
      Integer tmpInt = new Integer(i);
      seed = new byte[] {1, 2, 3, 4, 5, 6, 7, 8, 9, tmpInt.byteValue()};
      curTime = System.currentTimeMillis();
      mySpeed.generateKeys(seed);
      curTime = System.currentTimeMillis() - curTime;
      System.out.println(curTime + " ms");
    }
    System.out.println();  /**/

    System.gc();
    // key generation II
    paramM = 102;
    System.out.println("n=" + paramM);
    System.out.println("r=1");
    System.out.println("Key Generation with polynomial interpolation [QuartzLight, E=GF(2^" + extensionDegree +  ")]");
    for (int i = 0; i < paramM; i++) {
      Integer tmpInt = new Integer(i);
      seed = new byte[] {1, 2, 3, 4, 5, 6, 7, 8, 9, tmpInt.byteValue()};
      curTime = System.currentTimeMillis();
      mySpeed.generateKeysInterpoly(seed);
      curTime = System.currentTimeMillis() - curTime;
      System.out.println(curTime + " ms");
    }
    System.out.println();

    System.gc();
    // signature
    paramM = 102;
    System.out.println("n=" + paramM);
    System.out.println("r=1");
    System.out.println("Signature [QuartzLight, E=GF(2^" + extensionDegree +  ")]");
    int[][] manySig = new int[paramM][signatureLength];
    int[][] sigVec = new int[paramM][messageSigLength];
    for (int i = 0; i < paramM; i++) {
      sigVec[i] = new int[messageSigLength];
      for (int j = 0; j < sigVec[i].length; j++)
        sigVec[i][j] = Field2.rnd(rnd);
      curTime = System.currentTimeMillis();
      mySpeed.genSign(sigVec[i],manySig[i]);
      curTime = System.currentTimeMillis() - curTime;
      System.out.println(curTime + " ms");
    }
    System.out.println();

    System.gc();
    // signature verification
    repeats = 100;
    System.out.println("n=" + paramM);
    System.out.println("r=" + repeats);
    System.out.println("Verification [QuartzLight, E=GF(2^" + extensionDegree +  ")]");
    for (int i = 0; i < paramM; i++) {
      curTime = System.currentTimeMillis();
      for (int j = 0; j < repeats; j++) {
        mySpeed.valSign(sigVec[i],manySig[i]);
      }
      curTime = System.currentTimeMillis() - curTime;
      System.out.println(curTime + " ms");
    }
    System.out.println();

    System.out.println("n=0");  // finis!
    System.out.println();
    long overall2 = System.currentTimeMillis() - overallTime;
    System.out.println("Overall time: " + overall2 + " ms (about " + (((float)overall2) / (1000*60))+
                       " min, i.e., " +(((float)overall2) / (1000*60*60))  + " hours)");
  }

  /** Tests everything in this class
      @see #testIt
  */
  public static void main(String[] args) {
    // init static variables in our classes
    Multi2Mul.initAll();
    Poly2_nMax.initAll();

    //testIt();
    speedIt();
  }
}
