package HFE;

import java.util.*;
import java.lang.Math;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing QuartzLight in Java
//-----------------------------------------------------------------------------

/**
<p>
This class deals with multi-variable polynomials and
a general modulus provied to the constructor.
Basicly, this class calculates multiples of two polynomials
and one polynomial and one element of (GF(2))^n.
All calculations are done in GF(2) ({@link HFE.Field2})
Every entry is a polynomial of the form
f_i(x1,x2,...) = g^i*(b_0 + b_1*x_1 + b_2*x_2 + ... + b_n*x_n)
The reduction is done in the variable g.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.15
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


class Multi2Mul {

  /** Stores all the reduction steps.
      reduce[0] tells how to reduce the first polynomial,
      reduce[1] who to reduce the second polynomial, ...
    @serial 3D array
    @version 17.12.01
  */
  private Poly2Max[] reduce;

  /** Initializes the internal field {@link #reduce}
      @param modulus  polynomial over GF(2) for reduction
                      the number of variables must be 3 for our purpose
  */
  Multi2Mul(Poly2Max modulus) {
    //Check.assert(modulus.degree() == HFE2_n.extensionDegree);

    int degree = HFE2_n.extensionDegree; // number of variables in the multivariant polynomials we deal with
    int entryNum = (2*degree)-1;      // how many entries do we need?

    // init reduce array
    reduce = new Poly2Max[entryNum];
    reduce[0] = new Poly2Max(new int[] {0});  // our first entry is the constant polynomial 1
    for (int i = 1; i < entryNum; i++) {
      reduce[i] = new Poly2Max(reduce[i-1],1);  // use the last polynomial and multiply it by x using the shift-constructor
      reduce[i].mod(modulus);
    }
  }

  /** Computes res = a * b
      in terms of vector multiplication of polynomials. Uses {@link #reduce} to
      take the modulus into account.
  */
  public Poly2MultiSquare[] mul(Poly2MultiLinear[] a, Poly2MultiLinear[] b) {
    final int dimension = HFE2_n.extensionDegree;
    //Check.assert(dimension == a.length);
    //Check.assert(dimension == b.length);

    // get one temporary register
    Poly2MultiSquare tmp = new Poly2MultiSquare(dimension);

    // init our result vector
    Poly2MultiSquare[] res = new Poly2MultiSquare[dimension];
    for (int i = 0; i < dimension; i++)
      res[i] = new Poly2MultiSquare(dimension);

    // compute the result using the reduction polynomial
    // we go throuth the matrix from left upper to right lower
    // in diagonal manner. All results with the same g-power
    // are done in one step
    for (int i = 0; i < reduce.length; i++) {
      // init our start values
      int startValueB = Math.max(0,i-dimension+1);
      int aVal = Math.min(i,dimension-1);
      int bVal = startValueB;
      // get all values in diagonal manner
      tmp.setZero();
      while (aVal >= startValueB) {
        //Check.assert(aVal+bVal==i);
        tmp.mulAdd(a[aVal],b[bVal]);
        // now produce a new aVal/bVal-pair
        aVal--;
        bVal++;
      }
      // now reduce them
      for (int j = 0; j < dimension; j++) {
        if (reduce[i].coeff(j) != 0)
          res[j].mulAdd(reduce[i].coeff(j),tmp);
      }
    }

    return res;
  }


  /** Computes res = con * a * b
      in terms of vector multiplication of polynomials. Uses {@link #reduce} to
      take the modulus into account.
    @param con constants as a vector, i.e. a vector of polynomials of degree 0 or -1
    @param a linear polynomial in dimension variables
    @param b linear polynimial in dimension variables
  */
  public Poly2MultiSquare[] mul(int[] con, Poly2MultiLinear[] a, Poly2MultiLinear[] b) {
    final int dimension = HFE2_n.extensionDegree;
    //Check.assert(dimension == a.length, "a has wrong length, namely " + a.length);
    //Check.assert(dimension == b.length, "b has wrong length, namely " + b.length);
    //Check.assert(dimension == con.length, "con has wrong length, namely " + con.length);
    //Check.assert(Field2.isElement(con), "con is not over Field2.");

    // get temporary registers
    Poly2MultiSquare tmpS = new Poly2MultiSquare(dimension);
    Poly2MultiLinear tmpL = new Poly2MultiLinear(dimension);

    // init our first result vector
    Poly2MultiLinear[] resL = new Poly2MultiLinear[dimension];
    for (int i = 0; i < dimension; i++)
      resL[i] = new Poly2MultiLinear(dimension);

    // compute the result using the reduction polynomial
    // we go throuth the matrix from left upper to right lower
    // in diagonal manner. All results with the same g-power
    // are done in one step
    for (int i = 0; i < reduce.length; i++) {
      // init our start values
      int startValueB = Math.max(0,i-dimension+1);
      int aVal = Math.min(i,dimension-1);
      int bVal = startValueB;
      // get all values in diagonal manner
      tmpL.setZero();
      while (aVal >= startValueB) {
        //Check.assert(aVal+bVal==i);
        if (con[bVal] != 0)
          tmpL.add(a[aVal]);  // as con[bVal] is in GF(2), we have to multiply by 1!
        // now produce a new aVal/bVal-pair
        aVal--;
        bVal++;
      }
      // now reduce them
      for (int j = 0; j < dimension; j++) {
        if (reduce[i].coeff(j) != 0)
          resL[j].add(tmpL);  // as we are in GF(2), we have to multiply by 1!
      }
    }

    // init our second result vector
    Poly2MultiSquare[] resS = new Poly2MultiSquare[dimension];
    for (int i = 0; i < dimension; i++)
      resS[i] = new Poly2MultiSquare(dimension);

    // compute the result using the reduction polynomial
    // we go throuth the matrix from left upper to right lower
    // in diagonal manner. All results with the same g-power
    // are done in one step
    for (int i = 0; i < reduce.length; i++) {
      // init our start values
      int startValueB = Math.max(0,i-dimension+1);
      int aVal = Math.min(i,dimension-1);
      int bVal = startValueB;
      // get all values in diagonal manner
      tmpS.setZero();
      while (aVal >= startValueB) {
        //Check.assert(aVal+bVal==i);
        tmpS.mulAdd(resL[aVal],b[bVal]);
        // now produce a new aVal/bVal-pair
        aVal--;
        bVal++;
      }
      // now reduce them
      for (int j = 0; j < dimension; j++) {
        if (reduce[i].coeff(j) != 0)
          resS[j].mulAdd(reduce[i].coeff(j),tmpS);
      }
    }

    return resS;
  }

  /** Init method for this class
  */
  static public void initAll() {
    for (int i = 0; i < HFE2_n.extensionDegree; i++)
      mul_resL[i] = new Poly2MultiLinear(HFE2_n.extensionDegree);
  }

  /** Holds a static variable for the sake of speed
      which is used in mul.
  */
  static private Poly2MultiSquare mul_tmpS = new Poly2MultiSquare(HFE2_n.extensionDegree);

  /** Holds a static variable for the sake of speed
      which is used in mul.
  */
  static private Poly2MultiLinear mul_tmpL = new Poly2MultiLinear(HFE2_n.extensionDegree);

  /** Holds a static variable for the sake of speed
      which is used in mul.
  */
  static private Poly2MultiLinear[] mul_resL = new Poly2MultiLinear[HFE2_n.extensionDegree];

  /** Computes res = con * a * b
      in terms of vector multiplication of polynomials. Uses {@link #reduce} to
      take the modulus into account.
    @param res vector for our result polynomials
    @param con constants as a vector, i.e. a vector of polynomials of degree 0 or -1
    @param a linear polynomial in dimension variables
    @param b linear polynimial in dimension variables
  */
  void mul(Poly2MultiSquare[] res, int[] con, Poly2MultiLinear[] a, Poly2MultiLinear[] b) {
    final int dimension = HFE2_n.extensionDegree;
    //Check.assert(dimension == a.length, "a has wrong length, namely " + a.length);
    //Check.assert(dimension == b.length, "b has wrong length, namely " + b.length);
    //Check.assert(dimension == con.length, "con has wrong length, namely " + con.length);
    //Check.assert(Field2.isElement(con), "con is not over Field2.");
    //Check.assert(dimension == res.length, "res has wrong length, namely " + res.length);

    // reset our first temporary register
    for (int i = 0; i < dimension; i++)
      mul_resL[i].setZero();

    // compute the result using the reduction polynomial
    // we go throuth the matrix from left upper to right lower
    // in diagonal manner. All results with the same g-power
    // are done in one step
    for (int i = 0; i < reduce.length; i++) {
      // init our start values
      int startValueB = Math.max(0,i-dimension+1);
      int aVal = Math.min(i,dimension-1);
      int bVal = startValueB;
      // get all values in diagonal manner
      mul_tmpL.setZero();
      while (aVal >= startValueB) {
        //Check.assert(aVal+bVal==i);
        if (con[bVal] != 0)
          mul_tmpL.add(a[aVal]);  // as con[bVal] is in GF(2), we have to multiply by 1!
        // now produce a new aVal/bVal-pair
        aVal--;
        bVal++;
      }
      // now reduce them
      for (int j = 0; j < dimension; j++) {
        if (reduce[i].coeff(j) != 0)
          mul_resL[j].add(mul_tmpL);  // as we are in GF(2), we have to multiply by 1!
      }
    }

    // init our second result vector
    for (int i = 0; i < dimension; i++)
      res[i].setZero();

    // compute the result using the reduction polynomial
    // we go throuth the matrix from left upper to right lower
    // in diagonal manner. All results with the same g-power
    // are done in one step
    for (int i = 0; i < reduce.length; i++) {
      // init our start values
      int startValueB = Math.max(0,i-dimension+1);
      int aVal = Math.min(i,dimension-1);
      int bVal = startValueB;
      // get all values in diagonal manner
      mul_tmpS.setZero();
      while (aVal >= startValueB) {
        //Check.assert(aVal+bVal==i);
        mul_tmpS.mulAdd(mul_resL[aVal],b[bVal]);
        // now produce a new aVal/bVal-pair
        aVal--;
        bVal++;
      }
      // now reduce them
      for (int j = 0; j < dimension; j++) {
        if (reduce[i].coeff(j) != 0)
          res[j].mulAdd(reduce[i].coeff(j),mul_tmpS);
      }
    }
  }


  /** Computes res = con * poly
      in terms of vector multiplication of polynomials. Uses {@link #reduce} to
      take the modulus into account.
    @param con constants as a vector, i.e. a vector of polynomials of degree 0 or -1
    @param poly linear polynomial in dimension variables
  */
  public Poly2MultiLinear[] mul(int[] con, Poly2MultiLinear[] poly) {
    final int dimension = HFE2_n.extensionDegree;
    //Check.assert(dimension == poly.length, "poly has wrong length, namely " + poly.length);
    //Check.assert(dimension == con.length, "con has wrong length, namely " + con.length);
    //Check.assert(Field2.isElement(con), "con is not over Field2.");

    // get temporary registers
    Poly2MultiLinear tmp = new Poly2MultiLinear(dimension);

    // init our first result vector
    Poly2MultiLinear[] res = new Poly2MultiLinear[dimension];
    for (int i = 0; i < dimension; i++)
      res[i] = new Poly2MultiLinear(dimension);

    // compute the result using the reduction polynomial
    // we go throuth the matrix from left upper to right lower
    // in diagonal manner. All results with the same g-power
    // are done in one step
    for (int i = 0; i < reduce.length; i++) {
      // init our start values
      int startValueB = Math.max(0,i-dimension+1);
      int aVal = Math.min(i,dimension-1);
      int bVal = startValueB;
      // get all values in diagonal manner
      tmp.setZero();
      while (aVal >= startValueB) {
        //Check.assert(aVal+bVal==i);
        if (con[bVal] != 0)
          tmp.add(poly[aVal]); // as we are in GF(2), con[bVal] must be 1
        // now produce a new aVal/bVal-pair
        aVal--;
        bVal++;
      }
      // now reduce them
      for (int j = 0; j < dimension; j++) {
        if (reduce[i].coeff(j) != 0)
          res[j].add(tmp); // as we are in GF(2), reduce[i].coeff(j) must be 1 now!
      }
    }

    return res;
  }

  /** Computes res = con * poly
      in terms of vector multiplication of polynomials. Uses {@link #reduce} to
      take the modulus into account.
    @param res linear polynomials in dimension variables
    @param con constants as a vector, i.e. a vector of polynomials of degree 0 or -1
    @param poly linear polynomials in dimension variables
  */
  void mul(Poly2MultiLinear[] res, int[] con, Poly2MultiLinear[] poly) {
    final int dimension = HFE2_n.extensionDegree;
    //Check.assert(dimension == res.length, "res has wrong length, namely " + res.length);
    //Check.assert(dimension == poly.length, "poly has wrong length, namely " + poly.length);
    //Check.assert(dimension == con.length, "con has wrong length, namely " + con.length);
    //Check.assert(Field2.isElement(con), "con is not over Field2.");

    // init our result with zero
    for (int i = 0; i < dimension; i++)
      res[i].setZero();

    // compute the result using the reduction polynomial
    // we go throuth the matrix from left upper to right lower
    // in diagonal manner. All results with the same g-power
    // are done in one step
    for (int i = 0; i < reduce.length; i++) {
      // init our start values
      int startValueB = Math.max(0,i-dimension+1);
      int aVal = Math.min(i,dimension-1);
      int bVal = startValueB;
      // get all values in diagonal manner
      mul_tmpL.setZero();
      while (aVal >= startValueB) {
        //Check.assert(aVal+bVal==i);
        if (con[bVal] != 0)
          mul_tmpL.add(poly[aVal]); // as we are in GF(2), con[bVal] must be 1
        // now produce a new aVal/bVal-pair
        aVal--;
        bVal++;
      }
      // now reduce them
      for (int j = 0; j < dimension; j++) {
        if (reduce[i].coeff(j) != 0)
          res[j].add(mul_tmpL); // as we are in GF(2), reduce[i].coeff(j) must be 1 now!
      }
    }
  }


  /** Computes res = a * b
      in terms of vector multiplication of polynomials. Uses {@link #reduce} to
      take the modulus into account. This is an independent implementation to
      {@link #mul(Poly2MultiLinear[], Poly2MultiLinear[])} and regarded to be slower.
  */
  public Poly2MultiSquare[] mul2(Poly2MultiLinear[] a, Poly2MultiLinear[] b) {
    final int dimension = HFE2_n.extensionDegree;
    //Check.assert(dimension == a.length, "a has wrong length, namely " + a.length);
    //Check.assert(dimension == b.length, "b has wrong length, namely " + b.length);

    // get one temporary register
    Poly2MultiSquare tmp = new Poly2MultiSquare(dimension);

    // init our result vector
    Poly2MultiSquare[] res = new Poly2MultiSquare[dimension];
    for (int i = 0; i < dimension; i++)
      res[i] = new Poly2MultiSquare(dimension);

    // compute the result using the reduction polynomial
    // first take care about the elements on the diagonal
    for (int i = 0; i < dimension; i++) {
      tmp.mul(a[i],b[i]);
      for (int j = 0; j < dimension; j++) {
        if (reduce[2*i].coeff(j) != 0)
          res[j].mulAdd(reduce[2*i].coeff(j),tmp);
      }
    }
    // now compute the non-diagonal elements
    for (int i = 0; i < (dimension-1); i++) {
      for (int j = i+1; j < dimension; j++) {
        tmp.mul(a[i],b[j]);
        tmp.mulAdd(a[j],b[i]);
        for (int k = 0; k < dimension; k++) {
          if (reduce[i+j].coeff(k) != 0)
            res[k].mulAdd(reduce[i+j].coeff(k),tmp);
        }
      }
    }

    return res;
  }


  /** Computes res = con * a * b
      in terms of vector multiplication of polynomials. Uses {@link #reduce} to
      take the modulus into account. This is an independent implementation to
      {@link #mul(int[], Poly2MultiLinear[], Poly2MultiLinear[])} and regarded to be slower.
  */
  public Poly2MultiSquare[] mul2(int[] con, Poly2MultiLinear[] a, Poly2MultiLinear[] b) {
    final int dimension = HFE2_n.extensionDegree;
    //Check.assert(dimension == a.length, "a has wrong length, namely " + a.length);
    //Check.assert(dimension == b.length, "b has wrong length, namely " + b.length);
    //Check.assert(dimension == con.length, "con has wrong length, namely " + con.length);
    //Check.assert(Field2.isElement(con), "con is not over Field2.");

    // get one temporary register
    Poly2MultiSquare tmp = new Poly2MultiSquare(dimension);

    // init our result vector
    Poly2MultiSquare[] res1 = new Poly2MultiSquare[dimension];
    for (int i = 0; i < dimension; i++)
      res1[i] = new Poly2MultiSquare(dimension);

    // compute the result using the reduction polynomial
    // first take care about the elements on the diagonal
    for (int i = 0; i < dimension; i++) {
      tmp.mul(a[i],b[i]);
      for (int j = 0; j < dimension; j++) {
        if (reduce[2*i].coeff(j) != 0)
          res1[j].mulAdd(reduce[2*i].coeff(j),tmp);
      }
    }
    // now compute the non-diagonal elements
    for (int i = 0; i < (dimension-1); i++) {
      for (int j = i+1; j < dimension; j++) {
        tmp.mul(a[i],b[j]);
        tmp.mulAdd(a[j],b[i]);
        for (int k = 0; k < dimension; k++) {
          if (reduce[i+j].coeff(k) != 0)
            res1[k].mulAdd(reduce[i+j].coeff(k),tmp);
        }
      }
    }

    // the last step is to compute the result
    // in terms of the constant polynomials
    // init our final result vector
    Poly2MultiSquare[] res2 = new Poly2MultiSquare[dimension];
    for (int i = 0; i < dimension; i++)
      res2[i] = new Poly2MultiSquare(dimension);

    // compute the result using the reduction polynomial
    // first take care about the elements on the diagonal
    for (int i = 0; i < dimension; i++) {
      tmp.setZero();
      tmp.mulAdd(con[i],res1[i]);
      for (int j = 0; j < dimension; j++) {
        if (reduce[2*i].coeff(j) != 0)
          res2[j].mulAdd(reduce[2*i].coeff(j),tmp);
      }
    }
    // now compute the non-diagonal elements
    for (int i = 0; i < (dimension-1); i++) {
      for (int j = i+1; j < dimension; j++) {
        tmp.setZero();
        tmp.mulAdd(con[i],res1[j]);
        tmp.mulAdd(con[j],res1[i]);
        for (int k = 0; k < dimension; k++) {
          if (reduce[i+j].coeff(k) != 0)
            res2[k].mulAdd(reduce[i+j].coeff(k),tmp);
        }
      }
    }
    return res2;
  }

  /** Tests the functionality of the class.
      Is called by the main method.
  */
  public static void testIt () {
    System.out.println("start testing Multi2Mul ...");

    Random rnd = new Random();
    final int repeats = 10;   // how often do we run our tests
    final int variables = HFE2_n.extensionDegree;


    Multi2Mul testObj = new Multi2Mul(HFE2_n.generatingPolynomialAsMaxPoly);
    Poly2MultiLinear[] left = new Poly2MultiLinear[variables], right = new Poly2MultiLinear[variables],
                       degree0 = new Poly2MultiLinear[variables];
    Poly2MultiSquare[] res1 = new Poly2MultiSquare[variables], res2 = new Poly2MultiSquare[variables],
                       cmp = new Poly2MultiSquare[variables];
    int [] vec = new int[variables];

    // test if mul == mul2
    for (int i = 0; i < repeats; i++) {
      int[] a = new int[variables+1];
      int[] b = new int[variables+1];
      int fac = Field2.rnd(rnd);
      for (int k = 0; k < variables; k++) {
        for (int j = 0; j < a.length; j++) {
          a[j] = Field2.rnd(rnd);
          b[j] = Field2.rnd(rnd);
        }
        left[k] = new Poly2MultiLinear(a);
        right[k] = new Poly2MultiLinear(b);
      }
      cmp = testObj.mul(left,right);       // mul
      res1 = testObj.mul2(left,right);     // mul2
      res2 = testObj.mul2(right,left);
      for (int j = 0; j < variables; j++) {
        if (!cmp[j].isEqual(res1[j])) {
          System.out.println("Error 1.1 at " + j);
          System.out.print("Got:      "); res1[j].writeString(); System.out.println();
          System.out.print("Expected: ");  cmp[j].writeString(); System.out.println();
        }
        if (!cmp[j].isEqual(res2[j])) {
          System.out.println("Error 1.2 at " + j);
          System.out.print("Got:      "); res2[j].writeString(); System.out.println();
          System.out.print("Expected: ");  cmp[j].writeString(); System.out.println();
        }
      }
    }

    // try random values for mul and mul2
    for (int i = 0; i < repeats; i++) {
      int[] a = new int[variables+1];
      int[] b = new int[variables+1];
      for (int k = 0; k < variables; k++) {
        vec[k] = Field2.rnd(rnd);
        for (int j = 0; j < a.length; j++) {
          a[j] = Field2.rnd(rnd);
          b[j] = Field2.rnd(rnd);
        }
        left[k] = new Poly2MultiLinear(a);
        right[k] = new Poly2MultiLinear(b);
      }

      res1 = testObj.mul(vec,left,right);      // first mul
      res2 = testObj.mul2(vec,left,right);     // then mul2

      for (int j = 0; j < variables; j++) {
        if (!res1[j].isEqual(res2[j])) {
          System.out.println("Error 2.1 at " + j);
          System.out.print("Got:      "); res1[j].writeString(); System.out.println();
          System.out.print("Expected: ");  cmp[j].writeString(); System.out.println();
        }
      }
    }

    System.out.println("...finished testing.");
  }

  /** Tests how fast the package runs on the current hardware.
  */
  public final static void speedIt() {
    //Check.assert(false);  // make sure we do not have any assertion checkers
    long overallTime = System.currentTimeMillis();

    System.out.println("Start speed measurement Multi2Mul...");
    System.out.println("System parameter: ");
    System.out.println(" ExtensionDegree: " + HFE2_n.extensionDegree);
    Random rnd = new Random();

    Multi2Mul testObj = new Multi2Mul(HFE2_n.generatingPolynomialAsMaxPoly);
    final int variables = HFE2_n.extensionDegree;
    long curTime;
    int paramM;
    int repeats;
    Poly2MultiLinear[] left = new Poly2MultiLinear[variables], right = new Poly2MultiLinear[variables],
                       degree0 = new Poly2MultiLinear[variables];
    Poly2MultiSquare[] res1 = new Poly2MultiSquare[variables], res2 = new Poly2MultiSquare[variables],
                       cmp = new Poly2MultiSquare[variables];
    int [] vec = new int[variables];

    /**/
    System.gc();

    paramM = 12;
    repeats = 10;
    System.out.println("n=" + paramM);
    System.out.println("r=" + repeats);
    System.out.println("Construction");
    for (int i = 0; i < paramM; i++) {
      curTime = System.currentTimeMillis();
      for (int j = 0; j < repeats; j++) {
        testObj = new Multi2Mul(HFE2_n.generatingPolynomialAsMaxPoly);
      }
      curTime = System.currentTimeMillis() - curTime;
      System.out.println(curTime + " ms");
    }
    System.out.println();


    /**/
    System.gc();
    paramM = 12;
    repeats = 10;
    System.out.println("n=" + paramM);
    System.out.println("r=" + repeats);
    System.out.println("Multiplication");
    for (int mTime = 0; mTime < paramM; mTime++) {
      // generate random vectors
      for (int i = 0; i < repeats; i++) {
        int[] a = new int[variables+1];
        int[] b = new int[variables+1];
        for (int k = 0; k < variables; k++) {
          vec[k] = Field2.rnd(rnd);
          for (int j = 0; j < a.length; j++) {
            a[j] = Field2.rnd(rnd);
            b[j] = Field2.rnd(rnd);
          }
          left[k] = new Poly2MultiLinear(a);
          right[k] = new Poly2MultiLinear(b);
        }
      }

      curTime = System.currentTimeMillis();
      for (int j = 0; j < repeats; j++) {
        res1 = testObj.mul(vec,left,right);
      }
      curTime = System.currentTimeMillis() - curTime;
      System.out.println(curTime + " ms");
    }
    System.out.println();

    System.out.println("n=0");  // finis!
    System.out.println();
    long overall2 = System.currentTimeMillis() - overallTime;
    System.out.println("Overall time: " + overall2 + " ms (about " + (((float)overall2) / (1000*60))+
                       " min, i.e., " +(((float)overall2) / (1000*60*60))  + " hours)");
  }

  /** Calls the {@link #testIt} method
  */
  public static void main(String[] args) {
    //testIt();
    speedIt();
  }
}
