package HFE;

import java.util.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing QuartzLight in Java
//-----------------------------------------------------------------------------

/**
<p>
This class provides polynomials of degree one in n independent variables.
The polynomials are over the finite field GF(2) which is
provided in the class {@link Field2}.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.15
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.
 

final class Poly2MultiLinear {

    /** The number of variables in this polynomial.
      @serial should be greater than 0
    */
    private int varNum;

    /** Coefficients for the variables.
        The variables are x_1..x_n where
        n is {@link #varNum}. Position 0
        is for the constant term.
    */
    private int coeff[];


    /** sets all the coefficients to zero
        assumes that the polynomial has no variables
    */
    public Poly2MultiLinear() {
      varNum = 0;
      initMem();
    }

    /** Sets all the coefficients to zero by allocating new memory.
        Assumes that the polynomial has varNum variables.
       @param varNum number of variables for that polynomial
    */
    public Poly2MultiLinear(int varNum) {
      //Check.assert(varNum >= 0);
      this.varNum = varNum;
      initMem();
    }


    /** Creates a copy from the given polynomial.
        Uses memory copy for the coefficients.
    */
    public Poly2MultiLinear(final Poly2MultiLinear father) {
      varNum = father.varNum;
      coeff = (int[])father.coeff.clone();
    }


    /** {@link #array2poly}
    */
    public Poly2MultiLinear(final int[] inArray) {
      //Check.assert(inArray.length > 0);
      //Check.assert(Field2.isElement(inArray));
      varNum = inArray.length-1;
      array2poly(inArray);
    }

    /** Creates a new object by using inArray for the
        coefficients. Does no checking for arguments,
        no memory copy.
       @param varNum number of variables for this object
       @param inArray coefficients for the multivaribale polynomial
    */
    public Poly2MultiLinear(int varNum, final int[] inArray) {
      //Check.assert((varNum >= 0) && (varNum == (inArray.length-1)));
      //Check.assert(Field2.isElement(inArray));
      this.varNum = varNum;
      coeff = inArray;
    }

    /** Returns the coefficient for a given variable.
        Valid values: 0: constant term, 1..valNum: variable
    */
    public int coeff(int which) {
      //Check.assert((which >= 0) && (which <= varNum));
      return coeff[which];
    }

    /** returns the coefficient array
        without copying it.
    */
    public int[] coeffs() {
      return coeff;
    }

    /** returns the number of variables
        for that object
    */
    public int giveVarNum() {
      return varNum;
    }

    /** Compares this and whom coefficientwise.
        If the two polynomials have different number of variables they
        can not be equal.
      @return true iff all the coefficients are the same
              both objects
    */
    public boolean isEqual(Poly2MultiLinear whom) {
      if (whom.varNum != varNum) return false;
      for (int i = 0; i <= varNum; i++) {
        if (whom.coeff[i] != coeff[i]) return false;
      }
      return true;
    }

    /** Checks if the polynomial is the zero polynomial.
      @return true iff all coefficients are zero
    */
    public boolean isZero() {
      for (int i = 0; i <= varNum; i++)
        if (coeff[i] != 0) return false;
      return true;
    }


    /** Sets all the coefficients to zero
    */
    public void setZero() {
      for (int i = 0; i <= varNum; i++)
        coeff[i] = 0;
    }


    /** Allocates new memory for the coefficients.
        The coefficients are initalized with zero.
    */
    public void initMem() {
      coeff = new int[varNum+1];
    }


    /** Sets all the coefficients
        by the given array. Does neither memory
        copy nor argument checking.
    */
    public void setCoeffs(int [] inArray) {
      //Check.assert(varNum == (inArray.length-1));
      //Check.assert(Field2.isElement(inArray));
      coeff = inArray;
    }


    /** Sets the coefficients using the given coefficient.
        Copies the array and checks if the array
        has the right length.
       @throws IllegalArgumentException: array has the wrong length, needs to be (varNum+1)
    */
    public void array2poly(final int[] inArray) {
      //Check.assert(Field2.isElement(inArray));
      if (inArray.length != (varNum+1)) new IllegalArgumentException("Need " + varNum + " coefficients");
      coeff = (int[])inArray.clone();
    }


    /** calculates this = a + b
        as polynomial addition
    */
    public void add(final Poly2MultiLinear a, final Poly2MultiLinear b) {
      //Check.assert((a.varNum == varNum) && (b.varNum == varNum));
      for (int i = 0; i <= varNum; i++)
        coeff[i] = Field2.add(a.coeff[i],b.coeff[i]);
    }


    /** calculate this += whom
        as polynomial addition
    */
    public void add(final Poly2MultiLinear whom) {
      //Check.assert(whom.varNum == varNum);
      for (int i = 0; i <= varNum; i++)
        coeff[i] = Field2.add(whom.coeff[i],coeff[i]);
    }

    /** calculates this = a - b
        as polynomial subtraction
    */
    public void sub(final Poly2MultiLinear a, final Poly2MultiLinear b) {
      //Check.assert((a.varNum == varNum) && (b.varNum == varNum));
      for (int i = 0; i <= varNum; i++)
        coeff[i] = Field2.sub(a.coeff[i],b.coeff[i]);
    }


    /** calculate this -= whom
        as polynomial subtraction
    */
    public void sub(final Poly2MultiLinear whom) {
      //Check.assert(whom.varNum == varNum);
      for (int i = 0; i <= varNum; i++)
        coeff[i] = Field2.sub(coeff[i],whom.coeff[i]);
    }

    /** Evaluate the polynomial using the given value vector
        The entry value[0] is _not_ uses, value[1] is the first valid entry.
      @param values array with varNum+1 entries
      @return value of the current polynimial.
    */
    public int apply(int[] values) {
      //Check.assert(values.length == (varNum+1));
      //Check.assert(Field2.isElement(values));
      int res = coeff[0];
      for (int i = 1; i <= varNum; i++)
        res = Field2.add(res,Field2.mul(coeff[i],values[i]));

      return res;
    }

    /** Evaluate the polynomial using the given value vector
        and the given coefficient vector. The entry value[0] is _not_ uses,
        value[1] is the first valid entry.
      @param coeffs coefficients of a multivariable, linear polynomial
      @param values values for the variables of the polynomial, value[1] is the value for the first variable, ...
      @return value of the current polynimial.
    */
    public static int apply(int [] coeffs, int[] values) {
      //Check.assert(values.length == coeffs.length);
      //Check.assert(Field2.isElement(coeffs) && Field2.isElement(values));
      int res = coeffs[0];
      for (int i = 1; i < values.length; i++)
        res = Field2.add(res,Field2.mul(coeffs[i],values[i]));

      return res;
    }

    /** Writes the polynomial as a string to stdout
    */
    public void writeString() {
      for (int i = 1; i <= varNum; i++) {
        if (coeff[i] != 0) {
          if (coeff[i] != 1) System.out.print(coeff[i]);
          System.out.print("x_" + i + " + ");
        }
      }
      System.out.print(coeff[0]);
    }


    /** Tests the functionality for the whole class.
        Reports errors to stdout.
    */
    public static void testIt () {
      System.out.println("start testing Poly2MultiLinear ...");

      Random rnd = new Random();
      Poly2MultiLinear cur = new Poly2MultiLinear(), cur1 = new Poly2MultiLinear(),
                       cur2 = new Poly2MultiLinear(), cur3 = new Poly2MultiLinear();
      int curRnd, curRnd1, curRnd2, curRnd3, curRes;

      cur1 = new Poly2MultiLinear(new int[] {0,1,0});
      cur2 = new Poly2MultiLinear(2, new int[] {0,1,0});
      cur3 = new Poly2MultiLinear(new int[] {0,1,0,1});
      if ((!cur1.isEqual(cur2)) || (cur1.giveVarNum() != 2)) System.out.println("Error 1.1");
      if (cur1.isEqual(cur3) || (cur3.giveVarNum() != 3)) System.out.println("Error 1.2");
      cur1 = new Poly2MultiLinear();
      if (!cur1.isZero()) System.out.println("Error 1.3");
      cur2.setZero();
      if (!cur2.isZero()) System.out.println("Error 1.4");
      cur3 = new Poly2MultiLinear(3);
      if ((!cur3.isZero()) || (cur3.varNum != 3)) System.out.println("Error 1.5");
      cur3.setCoeffs(new int[] {0, 1, 1, 0});
      cur2 = new Poly2MultiLinear(cur3);
      if ((!cur2.isEqual(cur3)) || (cur2.varNum != 3)) System.out.println("Error 1.6");
      if ((cur3.coeff(0) != 0) || (cur3.coeff(1) != 1) ||
          (cur3.coeff(2) != 1) || (cur3.coeff(3) != 0)) System.out.println("Error 1.7");
      cur3.initMem();
      if ((!cur3.isZero()) || (cur3.varNum != 3)) System.out.println("Error 1.8");
      int [] c = cur2.coeffs();
      if ((c[0] != 0) || (c[1] != 1) ||
          (c[2] != 1) || (c[3] != 0) || (c.length != 4)) System.out.println("Error 1.9");

      cur1 = new Poly2MultiLinear(new int[] {0, 1, 1, 1});
      cur2 = new Poly2MultiLinear(new int[] {0, 1, 1, 1});
      cur3 = new Poly2MultiLinear(new int[] {0, 1, 0,1});
      cur3.add(cur1,cur2);
      if (!cur3.isZero()) System.out.println("Error 2.1");
      cur2.add(cur1);
      if (!cur2.isZero()) System.out.println("Error 2.2");

      cur1 = new Poly2MultiLinear(new int[] {0, 1, 1, 1});
      cur2 = new Poly2MultiLinear(new int[] {0, 1, 1, 1});
      cur3 = new Poly2MultiLinear(new int[] {0, 1, 0,1});
      cur3.sub(cur1,cur2);
      if (!cur3.isZero()) System.out.println("Error 3.1");
      cur2.sub(cur1);
      if (!cur2.isZero()) System.out.println("Error 3.2");

      cur1 = new Poly2MultiLinear(new int[] {0, 1, 0, 1});
      if (1 != cur1.apply(new int[] {0, 1, 1, 0})) System.out.println("Error 4.1");
      if (0 != cur1.apply(new int[] {1, 1, 0, 1})) System.out.println("Error 4.2");
      if (0 != cur1.apply(new int[] {0, 1, 0, 1}, new int[] {0, 1, 0, 1})) System.out.println("Error 4.3");
      if (1 != cur1.apply(new int[] {0, 1, 0, 0}, new int[] {0, 1, 0, 1})) System.out.println("Error 4.4");

      System.out.println("...finished testing.");
    }

    /** Mainly invoces the {@link #testIt} method.
    */
    public static void main(String[] args) {
        testIt();
    }
}
