package HFE;

import java.lang.*;
import java.io.*;
import java.util.*;
import java.security.*;
import sun.misc.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing QuartzLight in Java
//-----------------------------------------------------------------------------

/**
<p>
This class computes root for a given polynomial over GF(2^n).
The algorithm used in this class is described in
<center>
Lidl, Rudolf and Niederreiter, Harald:<br>
                 <i>Introduction to finite fields and their applications</i><br>
                 Cambridge University Press, 1986. ISBN 0-521-30706-6,<br>
                 pp. 150-156
</center>
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.15
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


class RootFinding implements CheckRoot {

  /** number of registers in this machine.
      This number is calculated as follows: 4 registers are needed as temporary registers during
      the programme run. The other 8 registers are the maximum number of registers which can
      be used for intermediate results:
      The Quartz polynomial has degree 129, each register used _must_ have degree 2 at most.
      As 2^8=256 (but 2^7=128), at most 8 polynomials of degree 2 can be found as factors of
      polynomial p. Therefore we obtain 12.
  */
  final static int regNum = 12;

  /** array of squares mod Poly. Must be initialized by initAll()
  */
  Poly2_nMax[] square = new Poly2_nMax[HFE2_n.extensionDegree+1];

  /** registers for root finding algorithm. Are initialized in initAll().
  */
  Poly2_nMax[] f = new Poly2_nMax[regNum];

  /** registers for root finding algorithm. Holds polynomial S
  */
  Poly2_nMax S = new Poly2_nMax();

  /** registers for root finding algorithm. Hold degrees of polynomials.
      These degrees must be set by hand!
  */
  int[] degF = new int[regNum];

  /** bitvector. If regUsed_0 is 1, f[0] is in use and so on.
  */
  int regUsed;

  /** temporary register for roots
  */
  Field2_n tmpRoot = new Field2_n();

  /** initializes all variables in this class
  */
  public RootFinding() {
    for (int i = 0; i < square.length; i++)
      square[i] = new Poly2_nMax();
    square[0].setCoeff(new Field2_n(1),1);  // set square[0] to be x
    for (int i = 0; i < regNum; i++)
      f[i] = new Poly2_nMax();
    regUsed = 0;
  }

  /** Method for register handling. Frees a given register.
  */
  private void setFree(int reg) {
    //Check.assert(reg < regUsed, "reg is too big: " + reg);
    //Check.assert((regUsed & (1 << reg)) != 0, "reg " + reg + " is not used " + regUsed);
    regUsed &= ~(1 << reg);
  }

  /** Method for register handling. Returns the name of a free register and
      allocates this register.
  */
  private int giveFree() {
    int tmp = regUsed;
    int i = 0;
    while ((tmp & 1) != 0) {
      i++;
      tmp >>= 1;
      //Check.assert(i < regNum, "i too big " + i + " with " + regUsed + " and " + tmp);
    }
    // after this step we know that i is the name of a free register

    // mark register as used
    regUsed |= 1 << i;

    return i;  // give name of register to caller
  }

  /** Method for register handling. Computes the degree of a given register
  */
  private void compDeg(int reg) {
    degF[reg] = f[reg].degree();
  }

  /** for debugging purposes only: returns if the root which was found is correct!
    @return did we find the correct root?
  */
  public boolean checkRoot(Field2_n root) {
    return false;
  }

  /** checks if S obtains a new root for f[0..regNum-1]
    @return true iff we do not search for more roots
  */
  private boolean checkRoot(CheckRoot cr) {
    int tmpReg = regUsed;
    int cur = 0;
    int gcd = giveFree();  // get a new register
    while (tmpReg != 0) {
      // register in use?
      if ((tmpReg & 1) != 0) {
//System.out.println("  cur: " + cur + " with deg " + degF[cur]);
        //Check.assert(degF[cur] > 1, "Degree of cur is too small: " + degF[cur]);
        // do we have a root in f?
        f[gcd].gcd(S,f[cur]);
        compDeg(gcd);
//System.out.println("    gcd: " + gcd + " with degree " + degF[gcd]);
        // is gcd of interest?
        //Check.assert(degF[gcd] == f[gcd].degree() && degF[cur] == f[cur].degree());
        if (((degF[gcd] != degF[cur]) && (degF[gcd] != 0))) {
//System.out.println("    try root finding");
          // seems so, therefore we need division
          int div = giveFree();
          f[div].div(f[cur],f[gcd]);
          compDeg(div);
//System.out.println("      div: " + div + " with degree " + degF[div]);
          setFree(cur);  // this register is of no use any more
          // no check if we got a new root
          //Check.assert(degF[gcd] == f[gcd].degree() && degF[div] == f[div].degree());
          if (degF[gcd] == 1) {
            // got a root for gcd
            //Check.assert(f[gcd].getCoeff(1).isEqual(new Field2_n(1)), "factor is not monic: " + f[gcd].stringIt());
            f[gcd].getCoeff(tmpRoot,0);
//System.out.println("!     found root for gcd: " + tmpRoot.stringIt());
            //setFree(gcd); gcd is assumed to be free, hence we do not need to set it free!
            if (cr.checkRoot(tmpRoot)) return true;
          } else gcd = giveFree();  // we need a new gcd
          if (degF[div] == 1) {
            // got a root for div
            //Check.assert(f[div].getCoeff(1).isEqual(new Field2_n(1)), "factor is not monic: " + f[div].stringIt());
            f[div].getCoeff(tmpRoot,0);
//System.out.println("!     found root for div: " + tmpRoot.stringIt());
            setFree(div);
            if (cr.checkRoot(tmpRoot)) return true;
          }
        }
      }
      // one more register...
      tmpReg >>= 1;
      cur++;
    }
    setFree(gcd);  // free gcd register
    return false; // we need another run
  }

  /** finds all roots of given polynomial poly
    @param ply Polynomial over GF(2^n) which roots are to be found
  */
  public boolean findRoots(Poly2_nMax poly, CheckRoot cr) {
    int polyDegree = poly.degree();
    // step one: compute x, x^2, x^4, ... in terms of mod Poly
    // assumes that square[0] is already initialized with x
    //Check.assert(square[0].degree() == 1, "square[0] has wrong degree: " + square[0].degree());
    for (int i = 1; i < HFE2_n.extensionDegree+1; i++)
      square[i].squareMod(square[i-1],poly,polyDegree);

    // step two: test if square[n] is x
    if (!square[0].isEqual(square[HFE2_n.extensionDegree])) {
      // as it isn't, we have to compute gcd(Poly, x^q+x)
      f[1].add(square[0],square[HFE2_n.extensionDegree]);
      f[0].gcd(poly,f[1]);
    } else {
      f[0].copy(poly);  // Poly = gcd(Poly,x^q-x) in this case
    }
    // after this step, f[0] holds a polynomial with monial linear factors _only_
    // hence f[0] can be exploit to get the roots of poly

    // moreover, if f[0] has degree 0, there is no root of poly
    compDeg(0);
    if (degF[0] <= 0) return false;  // there is no root at all!
    if (degF[0] == 1) { // in this case we are lucky: we have only one solution!
//System.out.println("found solution in the very first step!");
      //Check.assert(f[0].getCoeff(1).isEqual(new Field2_n(1)), "factor is not monic: " + f[0].stringIt());
      f[0].getCoeff(tmpRoot,0);
//System.out.println("!     found " + tmpRoot.stringIt());
      return cr.checkRoot(tmpRoot);
    }
    // initialize regUsed
    regUsed = 1; // only f[0] holds a valid entry at the beginning

    // compute S(x) = sum x^2^j for j=0, ..., n-1 using squares from above
    S.add(square[0],square[1]);
    for (int i = 2; i < HFE2_n.extensionDegree; i++)
      S.add(square[i]);

    // checks if there is a root for f[0] using S
    if (checkRoot(cr)) return true;
    if (regUsed == 0) return false;  // no more roots to check...
    S.addOne();
    if (checkRoot(cr)) return true;
    if (regUsed == 0) return false;  // no more roots to check...

    // now compute different values of beta, starting with beta = x
    Field2_n beta = new Field2_n(1);
    Field2_n beta_squares = new Field2_n();
    Field2_n tmp = new Field2_n();
    for (int i = 1; i < HFE2_n.extensionDegree; i++) {
      beta.mulX();  // multiply beta by x
      // compute S for this value of beta
      S.setZero();
      S.mulAdd(beta,square[0]);
      beta_squares.mul(beta,beta);
      for (int j = 1; j < HFE2_n.extensionDegree-1; j++) {
        S.mulAdd(beta_squares,square[j]);
        tmp.mul(beta_squares,beta_squares);
        beta_squares.copy(tmp);
      }
      S.mulAdd(beta_squares,square[HFE2_n.extensionDegree-1]);
      // after this step, we have a new S
      if (checkRoot(cr)) return true;
      if (regUsed == 0) return false;  // no more roots to check...
      S.addOne();
      if (checkRoot(cr)) return true;
      if (regUsed == 0) return false;  // no more roots to check...
    }
    return false;
  }

  /** Tests the functionality of the whole class.
      If there occurs any error it will be printed on stdout.
  */
  public static void testIt() {
    System.out.println("Start testing RootFinding...");

    RootFinding roots = new RootFinding();

    final int repeats = 10;
    Random rnd = new Random();
    Poly2_nMax cur, cur1, cur2, cur3;
    Field2_n coeff1, coeff2;
    cur = new Poly2_nMax();
    cur1 = new Poly2_nMax();
    cur2 = new Poly2_nMax();
    cur3 = new Poly2_nMax();

    // one solution, "hidden" in a square
    for (int i = 0; i < repeats; i++) {
      cur1 = new Poly2_nMax(rnd,0);  // construct polynomial of degree 1 which is monic
      cur1.setCoeff(new Field2_n(1),1);
      cur.mul(cur1,cur1);            // produce its square
      roots.findRoots(cur,roots);
      System.out.println("Run " + i);
    }

    // five solutions, hidden in a square
    for (int i = 0; i < repeats; i++) {
      cur1.setZero();
      cur1.addOne();
      for (int j = 0; j < 5; j++) {
        // compute one linear factor
        cur2 = new Poly2_nMax(rnd,0);  // construct polynomial of degree 1 which is monic
        cur2.setCoeff(new Field2_n(1),1);
        cur.mul(cur1,cur2);  // put cur2 _twice_ into cur1 (i.e. as a square)
        cur1.mul(cur,cur2);  // cur is a temporary register here
      }
      long curTime = System.currentTimeMillis();
      System.out.println("Run " + i + " started at " + curTime);
      roots.findRoots(cur1,roots);
      long curTime2 = System.currentTimeMillis();
      System.out.println("Run " + i + " ended at " + curTime2 + " (took " + (curTime2-curTime) + "ms in total)");

    }

    for (int i = 0; i < repeats; i++) {
      cur1 = new Poly2_nMax(rnd);
      coeff1 = new Field2_n(1);
      cur1.setCoeff(coeff1,Poly2_nMax.maxLen-1);
      coeff1 = new Field2_n(rnd);
      System.out.println("random value: " + coeff1.stringIt());
      // compute the value for this random value
      coeff2 = cur1.applyPoly(coeff1);
      // compute cur1 += coeff2
      cur1.getCoeff(coeff1,0);
      coeff1.add(coeff2);
      cur1.setCoeff(coeff1,0);
      long curTime = System.currentTimeMillis();
      System.out.println("Run " + i + " started at " + curTime);
      roots.findRoots(cur1,roots);
      long curTime2 = System.currentTimeMillis();
      System.out.println("Run " + i + " ended at " + curTime2 + " (took " + (curTime2-curTime) + "ms in total)");
      System.out.println();
    }

    // tests register handling functionality
    roots.regUsed = 0;
    for (int i = 0; i < regNum; i++) {
      int reg = roots.giveFree();
      if (i != reg) System.out.println("Error 1.0 at " + i + " with " + reg);
    }
    for (int i = 0; i < regNum; i++)
      roots.setFree(i);
    if (roots.regUsed != 0) System.out.println("Error 1.1 with " + roots.regUsed);

    System.out.println("...end testing.");
  }

  /** Tests everything in this class
      @see #testIt
  */
  public static void main(String[] args) {
    Poly2_nMax.initAll();
    testIt();
  }
}
