package HFE;

import java.util.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing QuartzLight in Java
//-----------------------------------------------------------------------------

/**
<p>
This class provides a vector of polynomials of degree one in n independent variables.
The polynomials are over the finite field GF(2) each. As all these
polynomials are processed in parallel, the coefficients
are represented in terms of {@link Field2_n} each.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.15
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


final class SquarePolyVec {

    /** The number of variables in this polynomial.
      @serial is an integer
    */
    public static int varNum = HFE2_n.extensionDegree;

    /** The number of coefficients in this polynomial
      @serial is an integer
    */
    public static int coeffNum = 1 + ((varNum+1)*varNum)/2;  // only true over GF(2), otherwise +3 rather than +1

   /** The number of rows in this vector.
      @serial is an integer
    */
    public static int elemNum = HFE2_n.extensionDegree;

    /** Coefficients for the variables.
        The variables are x_1..x_n where
        n is {@link #coeffNum}. Position 0
        is for the constant terms.
    */
    private Field2_n coeff[];


    /** Initializes all coefficients to zero. The resulting polynomial vector is _not_ affine!
    */
    public SquarePolyVec() {
      initMem();
    }

    /** Creates a copy from the given polynomial.
        Uses memory copy for the coefficients.
    */
    public SquarePolyVec(final SquarePolyVec father) {
      initMem();
      copy(father);
    }

    /** Creates a copy from the given <b>affine</b> polynomial.
        Uses memory copy for the coefficients.
    */
    public SquarePolyVec(final AffinePolyVec father) {
      initMem();
      Field2_n[] fc = father.coeffs();
      for (int i = 0; i < fc.length; i++)
        coeff[i].copy(fc[i]);
    }

    /** Creates a new object as the product of two given affine polynomials
    */
    public SquarePolyVec(final AffinePolyVec fac1, final AffinePolyVec fac2) {
      initMem();
      mul(fac1,fac2);
    }

    /** Returns the coefficients for a given variable.
        Valid values: 0: constant term, 1..valNum: variable, varNum+1: var_1*var_2, ...
    */
    public Field2_n coeff(int which) {
      //Check.assert((which >= 0) && (which < coeffNum));
      return coeff[which];
    }

    /** returns the coefficient array
        without copying it.
    */
    public Field2_n[] coeffs() {
      return coeff;
    }

    /** Compares this and whom coefficientwise.
        If the two polynomials have different number of variables they
        can not be equal.
      @return true iff all the coefficients are the same
              both objects
    */
    public boolean isEqual(SquarePolyVec whom) {
      for (int i = 0; i < coeffNum; i++) {
        if (!whom.coeff[i].isEqual(coeff[i])) return false;
      }
      return true;
    }


    /** Sets all the coefficients to zero
    */
    public void setZero() {
      for (int i = 0; i < coeffNum; i++)
        coeff[i].setZero();
    }

    /** Allocates new memory for the coefficients.
        The coefficients are initalized with zero.
    */
    public void initMem() {
      coeff = new Field2_n[coeffNum];
      for (int i = 0; i < coeffNum; i++)
        coeff[i] = new Field2_n();
    }

    /** Initializes all coefficients with random bits
    */
    public void rnd(Random rnd) {
      for (int i = 0; i < coeffNum; i++)
        coeff[i].rnd(rnd);
    }

    /** Copies the coefficients from another polynomial vector
    */
    public void copy(SquarePolyVec father) {
      for (int i = 0; i < coeffNum; i++)
        coeff[i].copy(father.coeff[i]);
    }

    /** Sets all the coefficients
        by the given array. Does neither memory
        copy nor argument checking.
    */
    public void setCoeffs(Field2_n [] inArray) {
      //Check.assert(coeffNum == inArray.length);
      coeff = inArray;
    }


    /** Transforms an affine transformation into
        an affine polynomial vector.
       @param inTransform affine tranformation
    */
    public void transform2poly(Transform2 inTransform) {
      // start with the constant term
      coeff[0].setZero();
      int[] m = inTransform.m;
      for (int i = 0; i < elemNum; i++)
        if (m[i] != 0) coeff[0].setBit(i);
      int[][] Ms = inTransform.M.getArray();
      for (int i = 0; i < elemNum; i++) {
        coeff[i+1].setZero();
        for (int j = 0; j < elemNum; j++)
          if (Ms[j][i] != 0) coeff[i+1].setBit(j);
      }
    }

    /** calculates this = a + b
        as polynomial addition
    */
    public void add(final SquarePolyVec a, final SquarePolyVec b) {
      for (int i = 0; i < coeffNum; i++)
        coeff[i].add(a.coeff[i],b.coeff[i]);
    }

    /** calculate this += whom
        as polynomial addition
    */
    public void add(final AffinePolyVec whom) {
      Field2_n[] w = whom.coeffs();
      for (int i = 0; i <= varNum; i++)
        coeff[i].add(w[i]);
    }

    /** calculate this += whom
        as polynomial addition
    */
    public void add(final SquarePolyVec whom) {
      for (int i = 0; i < coeffNum; i++)
        coeff[i].add(whom.coeff[i]);
    }

    /** calculates this = a - b
        as polynomial subtraction
    */
    public void sub(final SquarePolyVec a, final SquarePolyVec b) {
      for (int i = 0; i < coeffNum; i++)
        coeff[i].sub(a.coeff[i],b.coeff[i]);
    }


    /** calculate this -= whom
        as polynomial subtraction
    */
    public void sub(final SquarePolyVec whom) {
      for (int i = 0; i < coeffNum; i++)
        coeff[i].sub(whom.coeff[i]);
    }

    /** Evaluate the polynomial using the given value vector
        The entry value[0] is _not_ uses, value[1] is the first valid entry.
      @param res value of the current polynimial vector.
      @param values array with varNum+1 entries
    */
    public void apply(Field2_n res, int[] values) {
      //Check.assert(values.length == (varNum+1));
      //Check.assert(Field2.isElement(values));
      // constant term
      res.copy(coeff[0]);
      // linear terms
      for (int i = 1; i <= varNum; i++)
        if (values[i] == 1) res.add(coeff[i]);
      // quadratic terms
      int j = 1, k = 1;
      for (int i = varNum+1; i < coeffNum; i++) {
        k++; if (k > varNum) { j++; k=j+1; }
        //Check.assert(j <= varNum, "j is too big " + j);
        //Check.assert(k <= varNum, "k is too big " + k);
        if ((values[j] == 1) && (values[k] == 1)) res.add(coeff[i]);
      }
    }

    /** Evaluate the polynomial using the given value vector
        The entry value[0] is _not_ uses, value[1] is the first valid entry.
      @param res value of the current polynimial vector.
      @param value array with varNum entries as {@link Field2_n} element
    */
    public void apply(Field2_n res, Field2_n value) {
      // constant term
      res.copy(coeff[0]);
      // linear terms
      for (int i = 1; i <= varNum; i++)
        if (value.checkBit(i-1)) res.add(coeff[i]);
      // quadratic terms
      int j = 0, k = 0;
      for (int i = varNum+1; i < coeffNum; i++) {
        k++; if (k >= varNum) { j++; k=j+1; }
        //Check.assert(j < varNum, "j is too big " + j);
        //Check.assert(k < varNum, "k is too big " + k);
        if (value.checkBit(j) && value.checkBit(k)) res.add(coeff[i]);
      }

    }

    /** this *= fac
        in terms of quadratic polynomial vector multiplication
    */
    public void mul(Field2_n fac) {
      for (int i = 0; i < coeffNum; i++)
        coeff[i].mul(fac);
    }

    /** Static register for multiplication of two vectors
    */
    private static Field2_n mulTmp1 = new Field2_n();
    private static Field2_n mulTmp2 = new Field2_n();

    /** res = fac1*fac2
        in terms of affine polynomial vector multiplication
    */
    public void mul(final AffinePolyVec fac1, final AffinePolyVec fac2) {
      Field2_n[] a = fac1.coeffs();
      Field2_n[] b = fac2.coeffs();
      // assume that we are working on a (varNum+1)*(varNum+1) matrix
      // and reduce that to a triangle matrix for the terms which are
      // not on the diagonal
      // set the constant term
      coeff[0].mul(a[0],b[0]);
      // compute the linear coefficients
      for (int j = 1; j <= varNum; j++) {
        mulTmp1.add(a[0],a[j]);
        mulTmp2.add(b[0],b[j]);
        coeff[j].mul(mulTmp1,mulTmp2);
        coeff[j].sub(coeff[0]);  // subtract the linear term
      }
      // compute the quadratic terms
      int i = varNum+1;
      for (int k = 1; k <= varNum; k++) {
        for (int j = k+1; j <= varNum; j++) {
          mulTmp1.add(a[j],a[k]);
          mulTmp1.add(a[0]);
          mulTmp2.add(b[j],b[k]);
          mulTmp2.add(b[0]);
          coeff[i].mul(mulTmp1,mulTmp2);

          mulTmp1.add(coeff[j],coeff[k]);
          mulTmp1.sub(coeff[0]);
          coeff[i].sub(mulTmp1);
          i++;
        }
      }
    }

    /** res = fac1*fac2
        in terms of affine polynomial vector multiplication. Uses direct multiplication and is hence rather slow.
    */
    public void mulSlow(final AffinePolyVec fac1, final AffinePolyVec fac2) {
      Field2_n[] a = fac1.coeffs();
      Field2_n[] b = fac2.coeffs();
      // assume that we are working on a (varNum+1)*(varNum+1) matrix
      // and reduce that to a triangle matrix for the terms which are
      // not on the diagonal
      // set the constant term
      coeff[0].mul(a[0],b[0]);
      // set the linear coefficients
      for (int j = 1; j <= varNum; j++) {
        coeff[j].mul(a[0],b[j]);
        mulTmp1.mul(a[j],b[0]);
        coeff[j].add(mulTmp1);
      }
      int i = varNum+1;
      for (int k = 1; k <= varNum; k++) {
        mulTmp1.mul(a[k],b[k]);  // squares go to linear terms
        coeff[k].add(mulTmp1);
        for (int j = k+1; j <= varNum; j++) {
          // real quadratic terms
          coeff[i].mul(a[j],b[k]);
          mulTmp1.mul(a[k],b[j]);
          coeff[i].add(mulTmp1);
          i++;
        }
      }
    }

    /** Writes the polynomials as a string to stdout
    */
    public void writeString() {
      // linear terms
      for (int i = 1; i <= varNum; i++)
        if (!coeff[i].isZero()) System.out.print(coeff[i].stringIt()+ " x_" + i + " + ");
      // quadratic terms
      int j = 1, k = 1;
      for (int i = varNum+1; i < coeffNum; i++) {
        k++; if (k >= varNum) { j++; k=j+1; }
        //Check.assert(j < varNum, "j is too big " + j);
        //Check.assert(k < varNum, "k is too big " + k);
        if (!coeff[i].isZero()) System.out.print(coeff[i].stringIt()+ " x_" + j + "," + k + " + ");
      }
      System.out.print(coeff[0].stringIt());
    }


    /** Tests the functionality for the whole class.
        Reports errors to stdout.
    */
    public static void testIt () {
      System.out.println("start testing SquarePolyVec ...");

      final int repeats = 10;

      Random rnd = new Random();
      SquarePolyVec cur = new SquarePolyVec(), cur1 = new SquarePolyVec(),
                       cur2 = new SquarePolyVec(), cur3 = new SquarePolyVec();
      AffinePolyVec af1 = new AffinePolyVec(), af2 = new AffinePolyVec();
      Transform2 trans1 = new Transform2(rnd), trans2 = new Transform2(rnd);
      int curRnd, curRnd1, curRnd2, curRnd3, curRes;
      int [] testVec = new int[varNum], outVec = new int[varNum], valueVec = new int[varNum+1];
      Field2_n out1 = new Field2_n(), out2 = new Field2_n(), out3 = new Field2_n(),
               out4 = new Field2_n(), in1 = new Field2_n(), in2 = new Field2_n();

      // test apply
      for (int i = 0; i < repeats; i++) {
        in1.setZero();
        for (int j = 0; j < testVec.length; j++) {
          testVec[j] = Field2.rnd(rnd);
          valueVec[j+1] = testVec[j];
          if (testVec[j] == 1) in1.setBit(j);
        }
        af1 = new AffinePolyVec(trans1);
        cur1 = new SquarePolyVec(af1);
        cur2 = new SquarePolyVec(cur1);
        cur1.apply(out1,valueVec);
        af1.apply(out2,valueVec);
        if (!out2.isEqual(out1)) System.out.println("Error 1.1 with " + out1.stringIt() + " and " + out2.stringIt());
        cur2.apply(out2,in1);
        if (!out2.isEqual(out1)) System.out.println("Error 1.2 with " + out1.stringIt() + " and " + out2.stringIt());
        trans1.transform(testVec,outVec);
        out2.setElement(outVec);
        if (!out2.isEqual(out1)) System.out.println("Error 1.3 with " + out1.stringIt() + " and " + out2.stringIt());
      }

      // test multiplication
      for (int i = 0; i < repeats; i++) {
        af1.rnd(rnd);;
        af2 = af1.sqr();
        cur1.mul(af1,af1);  // should have be a square as result
        cur2 = new SquarePolyVec(af2);
        if (!cur1.isEqual(cur2))System.out.println("Error 1.4");

        af2.rnd(rnd);
        in1.rnd(rnd);
        cur1.mul(af1,af2);

        af1.apply(out1,in1);
        af2.apply(out2,in1);
        out1.mul(out2);
        cur1.apply(out2,in1);
        if (!out2.isEqual(out1)) System.out.println("Error 1.5 with " + out1.stringIt() + " and " + out2.stringIt());
      }

      // test multiplication further
      for (int i = 0; i < repeats; i++) {
        af1.rnd(rnd);;
        af2.rnd(rnd);
        in1.rnd(rnd);
        in2.rnd(rnd);
        cur1.mul(af1,af2);
        cur2.mul(af1,af2);
        cur2.mul(in1);
        cur1.apply(out1,in2);
        out1.mul(in1);
        cur2.apply(out2,in2);
        if (!out2.isEqual(out1)) System.out.println("Error 1.6 with " + out1.stringIt() + " and " + out2.stringIt());

        cur1.mul(af1,af2);
        cur2 = new SquarePolyVec(cur1);
        cur2.mul(in1);
        cur1.apply(out1,in2);
        out1.mul(in1);
        cur2.apply(out2,in2);
        if (!out2.isEqual(out1)) System.out.println("Error 1.7 with " + out1.stringIt() + " and " + out2.stringIt());
      }

      // test equality
      for (int i = 0; i < repeats; i++) {
        af1.rnd(rnd);;
        af2.rnd(rnd);
        in1.rnd(rnd);
        in2.rnd(rnd);
        cur1.mul(af1,af2);
        cur2 = new SquarePolyVec(cur1);
        if (!cur1.isEqual(cur2)) System.out.println("Error 1.8");
      }

      // test addition
      for (int i = 0; i < repeats; i++) {
        in1.rnd(rnd);
        cur1.rnd(rnd);
        cur2.rnd(rnd);
        cur3.add(cur1,cur2);
        cur1.apply(out1,in1);
        cur2.apply(out2,in1);
        cur3.apply(out3,in1);
        out4.add(out1,out2);

        if (!out4.isEqual(out3)) System.out.println("Error 1.9 with " + out3.stringIt() + " and " + out4.stringIt());
      }

      // test equality of multiplication
      for (int i = 0; i < repeats; i++) {
        af1.rnd(rnd);
        af2.rnd(rnd);
        cur1.mul(af1,af2);
        cur2.mulSlow(af1,af2);
        if (!cur1.isEqual(cur2)) System.out.println("Error 2.0");
      }

      // test componentwise
      af1.rnd(rnd);
      af2.rnd(rnd);
      cur1.mul(af1,af2);
      cur2.mulSlow(af1,af2);
      // check the constant term
      if (!cur1.coeff[0].isEqual(cur2.coeff[0])) System.out.println("constant not equal " + cur1.coeff[0].stringIt() + " and "  + cur2.coeff[0].stringIt());

      // linear terms
      for (int i = 1; i <= varNum; i++) {
        if (!cur1.coeff[i].isEqual(cur2.coeff[i])) System.out.println("linear not equal " + cur1.coeff[i].stringIt() + " and "  + cur2.coeff[i].stringIt());
      }

      // quadratic terms
      for (int i = varNum+1; i < coeffNum; i++) {
        if (!cur1.coeff[i].isEqual(cur2.coeff[i])) System.out.println("quadrat " + i +  " not equal " + cur1.coeff[i].stringIt() + " and "  + cur2.coeff[i].stringIt());
      }

      System.out.println("...finished testing.");
    }

    /** Tests how fast the package runs on the current hardware.
    */
    public final static void speedIt() {
      //Check.assert(false);  // make sure we do not have any assertion checkers
      long overallTime = System.currentTimeMillis();

      System.out.println("Start speed measurement SquarePolyVec...");
      System.out.println("System parameter: ");
      System.out.println(" VarNum: " + varNum);
      System.out.println(" CoeffNum: " + coeffNum);
      Random rnd = new Random();
      AffinePolyVec af1 = new AffinePolyVec(),af2 = new AffinePolyVec();
      SquarePolyVec sq1 = new SquarePolyVec(), sq2 = new SquarePolyVec();

      long curTime;
      int paramM;
      int repeats;

      /**/
      System.gc();
      // key generation
      paramM = 12;
      repeats = 100;
      System.out.println("n=" + paramM);
      System.out.println("r=" + repeats);
      System.out.println("Constructing from AffinePolyVec");
      for (int i = 0; i < paramM; i++) {
        af1.rnd(rnd);
        curTime = System.currentTimeMillis();
        for (int j = 0; j < repeats; j++)
          sq1 = new SquarePolyVec(af1);
        curTime = System.currentTimeMillis() - curTime;
        System.out.println(curTime + " ms");
      }
      System.out.println();

      System.gc();
      repeats = 10;
      System.out.println("n=" + paramM);
      System.out.println("r=" + repeats);
      System.out.println("Multiplication of two AffinePolyVec");
      for (int i = 0; i < paramM; i++) {
        af1.rnd(rnd);
        af2.rnd(rnd);
        curTime = System.currentTimeMillis();
        for (int j = 0; j < repeats; j++)
          sq1.mul(af1,af2);
        curTime = System.currentTimeMillis() - curTime;
        System.out.println(curTime + " ms");
      }
      System.out.println();

      System.gc();
      repeats = 10;
      System.out.println("n=" + paramM);
      System.out.println("r=" + repeats);
      System.out.println("Multiplication of two AffinePolyVec (slow)");
      for (int i = 0; i < paramM; i++) {
        af1.rnd(rnd);
        af2.rnd(rnd);
        curTime = System.currentTimeMillis();
        for (int j = 0; j < repeats; j++)
          sq1.mulSlow(af1,af2);
        curTime = System.currentTimeMillis() - curTime;
        System.out.println(curTime + " ms");
      }
      System.out.println();
      System.out.println("n=0");  // finis!
      System.out.println();
      long overall2 = System.currentTimeMillis() - overallTime;
      System.out.println("Overall time: " + overall2 + " ms (about " + (((float)overall2) / (1000*60))+
                         " min, i.e., " +(((float)overall2) / (1000*60*60))  + " hours)");
    }

    /** Mainly invoces the {@link #testIt} method.
    */
    public static void main(String[] args) {
        //testIt();
        speedIt();
    }
}
